/*  S-Lang console functions in C++
 *
 *  (C) Copyright 2004-2006 Denis Rojo <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <console.h>
#include <jutils.h>


///////////////////////////////////
// dirty signals

static bool screen_size_changed;
static void sigwinch_handler (int sig) {
  screen_size_changed = true;
  SLsignal (SIGWINCH, sigwinch_handler);
}

static bool keyboard_quit;
static void sigint_handler (int sig) {
  SLsignal_intr (SIGINT, sigint_handler);
  keyboard_quit = true;
#if SLANG_VERSION < 20000
  if (SLang_Ignore_User_Abort == 0) 
	  SLang_Error = USER_BREAK;
#endif
}

/* non blocking getkey */
static int getkey_handler() {
  unsigned int ch = 0;
  if(SLang_input_pending(0))
    //    return SLang_getkey();
    ch = SLang_getkey();
  //  if(ch) func("SLang_getkey in getkey_handler detected char %u",ch);
  return ch;
}

// end of dirty signals
/////////////////////////////////////


Console::Console() {

  buffer = NULL;

  w = h = 0;
  x = y = 0;

  wrap = false;
}

Console::~Console() { close(); }

bool Console::init() {

  //  setenv("TERM","xterm-color",0); 
  SLtt_get_terminfo();
  SLkp_init(); // keyboard interface

  SLang_init_tty(-1, 0, 0);

  SLsmg_init_smg(); // screen manager

  screen_size_changed = false;

  // register signals
  SLsignal (SIGWINCH, sigwinch_handler);
  SLang_set_abort_signal(sigint_handler);

  SLkp_set_getkey_function(getkey_handler);
  
  SLtt_set_cursor_visibility(1);

  // SLsmg_Tab_Width = 8;
  // SLsmg_Display_Eight_Bit = 1;
  // SLtt_Use_Ansi_Colors = 0;
  // SLtt_Term_Cannot_Scroll = 0;

  w = SLtt_Screen_Cols;
  h = SLtt_Screen_Rows;

  buffer = (char*)calloc((w*h)*2, sizeof(char));

  int c;
  for(c = 0; c< w*h; c++)
    buffer[c] = CHAR_BLANK; // space

  refresh();

  return true;
}

void Console::close() {

  if(buffer) {
    free(buffer);
    buffer = NULL;
  }

  SLsmg_reset_smg();
  SLang_reset_tty();
}

uint64_t Console::getkey() {
  // low level
  return SLkp_getkey();
}

void Console::putkey(uint64_t key) {
  // interprets a keycode and perform the action (or write a letter)
  switch(key) {

  case KEY_ENTER:
    /*** scroll_down
	 all rows under the cursor (excluding the cursor row) */
    if(y<h) {

      // move down everything below
      scroll_down(y+2, 1);

      // move down the current line
      move( 0, y+1, x, y, w-x );

      /* refresh all screen (debugging)
      for(int c=0; c<h; c++) {
	SLsmg_gotorc(c,0);
	SLsmg_write_nchars( &buffer[ c*w ], w);
	}*/

      y++;
      x = 0;
    }

    break;

  case KEY_TAB:
    if(x<w) x += 8;
    if(x>w) x = w;
    break;

  case KEY_BACKSPACE:
  case KEY_BACKSPACE_APPLE:
  case KEY_BACKSPACE_SOMETIMES:
    if( x < 1 ) return;
    if(x>0) {

      if( x >= w ) {
	// if at the end of the line just step one back
	buffer[ x + (y*w) ] = CHAR_BLANK;
      } else
	// move all right of the cursor to 1 step left
	move( x-1, y, x, y, w-x);
      
      x--; // we still must adjust coords
    }
    break;
    
  case SL_KEY_LEFT:
    if(x>0) x--;
    break;

  case SL_KEY_RIGHT:
    if(x<w) x++;
    break;

  case SL_KEY_UP:
    if(y>0) y--;
    break;

  case SL_KEY_DOWN:
    if(y<h) y++;
    break;

  case SL_KEY_PPAGE:
    y -= 20;
    if(y<0) y = 0;
    break;

  case SL_KEY_NPAGE:
    y += 20;
    if(y<0) y = 0;
    break;

    /////////////////////////////////////
    // some typical GNU/Emacs keybindings
  case KEY_CTRL_A:
  case KEY_HOME:
    x=0;
    break;

  case KEY_CTRL_E:
    x=w;
    break;

    
  default:
    if(x<w) {

      if( x >= w) return;

      /*** scroll_right
	   scroll all strings from the cursor (included) to the right */      
      move(x+1, y, x, y, w-x);

      putch(key);
    }

    break;
  }
}

/////////////
// INTERNALS
 ///////////

void Console::clean_overlap(int dest_off, int src_off, int len) {

  if(dest_off < src_off) { // moving backwards
    
    if(dest_off + len > src_off) { // overlap
      /* dst        src
                     __________
	 |==========|===
	             *** <- overlap

	 so we must clean
	 starting from src+overlap (= dst+len)
	 until len-overlap
      */
      memset( &buffer[ (dest_off + len) ], CHAR_BLANK,
	      len - ((dest_off + len) - src_off) );
      

    } else { // don't overlap

      memset( &buffer[ src_off ], CHAR_BLANK, len );

    }
    
  } else { // moving forward

    if(src_off + len > dest_off) { // overlap
      /* src             dst
                         _______________
	 |===============|===
	                  *** <- overlap

	 so we must clean
	 starting from src
	 until len-overlap (dst-src)
      */
      memset( &buffer[ src_off ], CHAR_BLANK,
	      dest_off - src_off);
	
      
    } else { // don't overlap

      memset( &buffer[ src_off ], CHAR_BLANK, len );

    }
  }
}


void Console::scroll_down(int y, int len) {

  // trivial checks
  if(y>h) return;

  int src_off = y*w;
  int dest_off = (y - len) *w;
  int size = (h - y - len) *w;

  memmove( &buffer[ src_off ], &buffer[ dest_off ], size);

  clean_overlap(src_off, dest_off, size);

  int c;

  //  for( c = y; c < h; c++)
  //    SLsmg_write_nchars( &buffer[ c * w ], w ); 

  // refresh all screen (debugging)
  for(c=0; c<h; c++) {
    SLsmg_gotorc(c,0);
    SLsmg_write_nchars( &buffer[ c*w ], w);
  }
  
  SLsmg_gotorc(y, x);

}



void Console::move(int dest_x, int dest_y, int src_x, int src_y, int len) {
  
  int dest_off = dest_x + (dest_y * w);
  int src_off  = src_x + (src_y * w);

  // trivial checks
  if(dest_off == src_off) return;
  if(dest_x > w) return;
  if(dest_y > h) return;

      
  if(!wrap) { // don't wrap, things fade out on the right
    
    /*     dest_x      w   len
      _______________________
      |    dddddddddddd|dddd     */
    if(dest_x + len > w)
      len = w - dest_x;

  }

  // clip len on end of screen
  if( dest_off + len > w*h )
    len = (w*h) - dest_off;
    

  // shift in buffer first
  memmove( &buffer[ dest_off ], &buffer[ src_off ], len);

  // blank the source
  clean_overlap(dest_off, src_off, len);



       
  // if is spanned across different rows, refresh all
  if(src_y != dest_y ) {
    int c;

    // refresh the src rows
    for(c=0; c < h-src_y; c++) {
      
      SLsmg_gotorc(src_y+c, 0);
      SLsmg_write_nchars( &buffer[ src_off - src_x + (c*w) ], w );
      if(len < (c*w)) break;

    }

    // refresh the dest rows
    for(c=0; c< h-dest_y; c++) {
      
      SLsmg_gotorc(dest_y+c, 0);
      SLsmg_write_nchars( &buffer[ dest_off - dest_x + (c*w) ], w );
      if(len < (c*w)) break;

    }

  } else { // move was on the same line

    SLsmg_gotorc(dest_y, 0);
    SLsmg_write_nchars( &buffer[ dest_off - dest_x ], w );

  }

  if(wrap) {
    
    /*         dest_x       w   len
	  __________________________
	  |    ddddddddddddd|xxxxxx
	  |dddddd
	  refresh also the line below dest if wrapping */
    if(dest_x + len > w) {
      SLsmg_gotorc(dest_y+1, 0);
      SLsmg_write_nchars( &buffer[ (dest_y+1) * w ], w );    
    }
  }

  // return at initial cursor position
  SLsmg_gotorc( y, x );
}



bool Console::refresh() {

  if(keyboard_quit) return false;

  /* S-Lang says: 
   * All well behaved applications should block signals that
   * may affect the display while performing screen update. */
  SLsig_block_signals ();
  
  if(screen_size_changed) {
    SLtt_get_screen_size ();
    SLsmg_reinit_smg ();

    Console::w = SLtt_Screen_Cols;
    Console::h = SLtt_Screen_Rows;

    screen_size_changed = false;
  }

  SLsmg_gotorc(y, x);

  SLsmg_refresh();
  
  SLsig_unblock_signals();

  return true;
}

void Console::putch(char ch) {
  SLsmg_write_char(ch);
  buffer[x+(w*y)] = ch;
  x++;
}

////////////////////////
/// messaging interface (TODO)
void Console::notice (char *str) { }
void Console::act    (char *str) { }
void Console::warning(char *str) { }
void Console::error  (char *str) { }
void Console::func   (char *str) { }
////////////////////////
